# -*- coding: UTF-8 -*-
import typing
from itertools import chain
from librouteros.protocol import (
    cast_to_api,
)
from librouteros.types import (
    QueryGen,
    ResponseIter,
)


class Key:

    def __init__(self, name: str):
        self.name = name

    def __eq__(self, other):
        yield f'?={self}={cast_to_api(other)}'

    def __ne__(self, other):
        yield from self == other
        yield '?#!'

    def __lt__(self, other):
        yield f'?<{self}={cast_to_api(other)}'

    def __gt__(self, other):
        yield f'?>{self}={cast_to_api(other)}'

    def __str__(self) -> str:
        return str(self.name)

    # pylint: disable=invalid-name
    def In(self, one, *elems):
        yield from self == one
        yield from chain.from_iterable(self == str(elem) for elem in elems)
        yield from ('?#|', ) * len(elems)


class Query:

    def __init__(self, path, keys: typing.Sequence[Key], api):
        self.path = path
        self.keys = keys
        self.api = api
        self.query: typing.Tuple[str, ...] = tuple()

    def where(self, *args: str):
        self.query = tuple(chain.from_iterable(args))
        return self

    def __iter__(self) -> ResponseIter:
        keys = ','.join(str(key) for key in self.keys)
        keys = f'=.proplist={keys}'
        cmd = str(self.path.join('print'))
        return iter(self.api.rawCmd(cmd, keys, *self.query))


def And(left: QueryGen, right: QueryGen, *rest: QueryGen) -> QueryGen:
    #pylint: disable=invalid-name
    yield from left
    yield from right
    yield from chain.from_iterable(rest)
    yield '?#&'
    yield from ('?#&', ) * len(rest)


def Or(left: QueryGen, right: QueryGen, *rest: QueryGen) -> QueryGen:
    #pylint: disable=invalid-name
    yield from left
    yield from right
    yield from chain.from_iterable(rest)
    yield '?#|'
    yield from ('?#|', ) * len(rest)



# -*- coding: UTF-8 -*-

import typing
from struct import pack, unpack #pylint: disable=no-name-in-module
from logging import getLogger, NullHandler

from librouteros.exceptions import (
    ProtocolError,
    FatalError,
)
from librouteros.connections import SocketTransport

LOGGER = getLogger('librouteros')
LOGGER.addHandler(NullHandler())


def parse_word(word: str) -> typing.Tuple[str, typing.Any]:
    """
    Split given attribute word to key, value pair.

    Values are casted to python equivalents.

    :param word: API word.
    :returns: Key, value pair.
    """
    mapping = {'yes': True, 'true': True, 'no': False, 'false': False}
    _, key, value = word.split('=', 2)
    try:
        value = int(value)                # type: ignore
    except ValueError:
        value = mapping.get(value, value) # type: ignore
    return (key, value)


def cast_to_api(value: typing.Any) -> str:
    """Cast python equivalent to API."""
    mapping = {True: 'yes', False: 'no'}
    # this is necesary because 1 == True, 0 == False
    if type(value) == int:
        return str(value)
    return mapping.get(value, str(value))


def compose_word(key: str, value: typing.Any) -> str:
    """
    Create a attribute word from key, value pair.
    Values are casted to api equivalents.
    """
    return f'={key}={cast_to_api(value)}'


class Encoder:

    def encodeSentence(self, *words: str) -> bytes:
        """
        Encode given sentence in API format.

        :param words: Words to endoce.
        :returns: Encoded sentence.
        """
        encoded = b''.join(self.encodeWord(word) for word in words)
        # append EOS (end of sentence) byte
        encoded += b'\x00'
        return encoded

    def encodeWord(self, word: str) -> bytes:
        """
        Encode word in API format.

        :param word: Word to encode.
        :returns: Encoded word.
        """
        #pylint: disable=no-member
        encoded_word = word.encode(encoding=self.encoding, errors='strict') # type: ignore
        return Encoder.encodeLength(len(word)) + encoded_word

    @staticmethod
    def encodeLength(length: int) -> bytes:
        """
        Encode given length in mikrotik format.

        :param length: Integer < 268435456.
        :returns: Encoded length.
        """
        if length < 128:
            ored_length = length
            offset = -1
        elif length < 16384:
            ored_length = length | 0x8000
            offset = -2
        elif length < 2097152:
            ored_length = length | 0xC00000
            offset = -3
        elif length < 268435456:
            ored_length = length | 0xE0000000
            offset = -4
        else:
            raise ProtocolError(f'Unable to encode length of {length}')

        return pack('!I', ored_length)[offset:]


class Decoder:

    @staticmethod
    def determineLength(length: bytes) -> int:
        """
        Given first read byte, determine how many more bytes
        needs to be known in order to get fully encoded length.

        :param length: First read byte.
        :return: How many bytes to read.
        """
        integer = ord(length)

        #pylint: disable=no-else-return
        if integer < 128:
            return 0
        elif integer < 192:
            return 1
        elif integer < 224:
            return 2
        elif integer < 240:
            return 3

        raise ProtocolError(f'Unknown controll byte {length!r}')

    @staticmethod
    def decodeLength(length: bytes) -> int:
        """
        Decode length based on given bytes.

        :param length: Bytes string to decode.
        :return: Decoded length.
        """
        bytes_length = len(length)

        if bytes_length < 2:
            offset = b'\x00\x00\x00'
            xor = 0
        elif bytes_length < 3:
            offset = b'\x00\x00'
            xor = 0x8000
        elif bytes_length < 4:
            offset = b'\x00'
            xor = 0xC00000
        elif bytes_length < 5:
            offset = b''
            xor = 0xE0000000
        else:
            raise ProtocolError(f'Unable to decode length of {length!r}')

        decoded: int = unpack('!I', (offset + length))[0]
        decoded ^= xor
        return decoded


class ApiProtocol(Encoder, Decoder):

    def __init__(self, transport: SocketTransport, encoding: str):
        self.transport = transport
        self.encoding = encoding

    @staticmethod
    def log(direction_string: str, *sentence: str) -> None:
        for word in sentence:
            LOGGER.debug(f'{direction_string} {word!r}')
        LOGGER.debug(f'{direction_string} EOS')

    def writeSentence(self, cmd: str, *words: str) -> None:
        """
        Write encoded sentence.

        :param cmd: Command word.
        :param words: Aditional words.
        """
        encoded = self.encodeSentence(cmd, *words)
        self.log('<---', cmd, *words)
        self.transport.write(encoded)

    def readSentence(self) -> typing.Tuple[str, typing.Tuple[str, ...]]:
        """
        Read every word untill empty word (NULL byte) is received.

        :return: Reply word, tuple with read words.
        """
        sentence = tuple(word for word in iter(self.readWord, ''))
        self.log('--->', *sentence)
        reply_word, words = sentence[0], sentence[1:]
        if reply_word == '!fatal':
            self.transport.close()
            raise FatalError(words[0])
        return reply_word, words

    def readWord(self) -> str:
        byte = self.transport.read(1)
        # Early return check for null byte
        if byte == b'\x00':
            return ''
        to_read = self.determineLength(byte)
        byte += self.transport.read(to_read)
        length = self.decodeLength(byte)
        word = self.transport.read(length)
        return word.decode(encoding=self.encoding, errors='ignore')

    def close(self) -> None:
        self.transport.close()



# -*- coding: UTF-8 -*-

import typing
from posixpath import join as pjoin
from librouteros.exceptions import TrapError, MultiTrapError
from librouteros.protocol import (
    compose_word,
    parse_word,
    ApiProtocol,
)
from librouteros import query

from librouteros.types import (
    ReplyDict,
    ResponseIter,
    Response,
)


class Api:

    def __init__(self, protocol: ApiProtocol):
        self.protocol = protocol

    def __call__(self, cmd: str, **kwargs: typing.Any) -> ResponseIter:
        """
        Call Api with given command.
        Yield each row.

        :param cmd: Command word. eg. /ip/address/print
        :param kwargs: Dictionary with optional arguments.
        """
        words = (compose_word(key, value) for key, value in kwargs.items())
        self.protocol.writeSentence(cmd, *words)
        yield from self.readResponse()

    def rawCmd(self, cmd: str, *words: str) -> ResponseIter:
        """
        Call Api with given command and raw words.
        End user is responsible to properly format each api word argument.
        :param cmd: Command word. eg. /ip/address/print
        :param args: Iterable with optional plain api arguments.
        """
        self.protocol.writeSentence(cmd, *words)
        yield from self.readResponse()

    def readSentence(self) -> typing.Tuple[str, ReplyDict]:
        """
        Read one sentence and parse words.

        :returns: Reply word, dict with attribute words.
        """
        reply_word, words = self.protocol.readSentence()
        return reply_word, dict(parse_word(word) for word in words)

    def readResponse(self) -> Response:
        """
        Yield each sentence untill !done is received.

        :throws TrapError: If one !trap is received.
        :throws MultiTrapError: If > 1 !trap is received.
        """
        traps = []
        reply_word = None
        response = []
        while reply_word != '!done':
            reply_word, words = self.readSentence()
            if reply_word == '!trap':
                traps.append(TrapError(**words))
            elif reply_word in ('!re', '!done') and words:
                response.append(words)

        if len(traps) > 1:
            raise MultiTrapError(*traps)
        if len(traps) == 1:
            raise traps[0]
        return response

    def close(self) -> None:
        self.protocol.close()

    def path(self, *path: str):
        return Path(
            path='',
            api=self,
        ).join(*path)


class Path:
    """Represents absolute command path."""

    def __init__(self, path: str, api: Api):
        self.path = path
        self.api = api

    def select(self, key: query.Key, *other: query.Key) -> query.Query:
        keys = (key, ) + other
        return query.Query(path=self, keys=keys, api=self.api)

    def __str__(self) -> str:
        return self.path

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self}>"

    def __iter__(self) -> ResponseIter:
        yield from self('print')

    def __call__(self, cmd: str, **kwargs: typing.Any) -> ResponseIter:
        yield from self.api(
            self.join(cmd).path,
            **kwargs,
        )

    def join(self, *path: str):
        """Join current path with one or more path strings."""
        return Path(
            api=self.api,
            path=pjoin('/', self.path, *path).rstrip('/'),
        )

    def remove(self, *ids: str) -> None:
        joined = ','.join(ids)
        tuple(self(
            'remove',
            **{'.id': joined},
        ))

    def add(self, **kwargs: typing.Any) -> str:
        ret = self(
            'add',
            **kwargs,
        )
        return tuple(ret)[0]['ret']

    def update(self, **kwargs: typing.Any) -> None:
        tuple(self(
            'set',
            **kwargs,
        ))



# -*- coding: UTF-8 -*-

import typing
from socket import create_connection
from collections import ChainMap

from librouteros.exceptions import (
    ConnectionClosed,
    FatalError,
)
from librouteros.connections import SocketTransport
from librouteros.protocol import ApiProtocol
from librouteros.login import (
    plain,
    token,
)
from librouteros.api import Api

DEFAULTS = {
    'timeout': 10,
    'port': 8728,
    'saddr': '',
    'subclass': Api,
    'encoding': 'ASCII',
    'ssl_wrapper': lambda sock: sock,
    'login_method': plain,
}


def connect(host: str, username: str, password: str, **kwargs) -> typing.Type[Api]:
    """
    Connect and login to routeros device.
    Upon success return a Api class.

    :param host: Hostname to connecto to. May be ipv4,ipv6,FQDN.
    :param username: Username to login with.
    :param password: Password to login with. Only ASCII characters allowed.
    :param timeout: Socket timeout. Defaults to 10.
    :param port: Destination port to be used. Defaults to 8728.
    :param saddr: Source address to bind to.
    :param subclass: Subclass of Api class. Defaults to Api class from library.
    :param ssl_wrapper: Callable (e.g. ssl.SSLContext instance) to wrap socket with.
    :param login_method: Callable with login method.
    """
    arguments = ChainMap(kwargs, DEFAULTS)
    transport = create_transport(host, **arguments)
    protocol = ApiProtocol(transport=transport, encoding=arguments['encoding'])
    api = arguments['subclass'](protocol=protocol)

    try:
        arguments['login_method'](api=api, username=username, password=password)
        return api
    except (ConnectionClosed, FatalError):
        transport.close()
        raise


def create_transport(host: str, **kwargs) -> SocketTransport:
    sock = create_connection((host, kwargs['port']), kwargs['timeout'], (kwargs['saddr'], 0))
    sock = kwargs['ssl_wrapper'](sock)
    return SocketTransport(sock=sock)



# -*- coding: UTF-8 -*-

import typing
from binascii import unhexlify, hexlify #pylint: disable=no-name-in-module
from hashlib import md5


def encode_password(token: str, password: str) -> str:
    #pylint: disable=redefined-outer-name
    token_bytes = token.encode('ascii', 'strict')
    token_bytes = unhexlify(token)
    password_bytes = password.encode('ascii', 'strict')
    hasher = md5()
    hasher.update(b'\x00' + password_bytes + token_bytes)
    password_bytes = hexlify(hasher.digest())
    return '00' + password_bytes.decode('ascii', 'strict')


def token(api: typing.Any, username: str, password: str) -> None:
    """Login using pre routeros 6.43 authorization method."""
    sentence = api('/login')
    tok = tuple(sentence)[0]['ret']
    encoded = encode_password(tok, password)
    tuple(api('/login', **{'name': username, 'response': encoded}))


def plain(api: typing.Any, username: str, password: str) -> None:
    """Login using post routeros 6.43 authorization method."""
    tuple(api('/login', **{'name': username, 'password': password}))



# -*- coding: UTF-8 -*-
from socket import socket
from librouteros.exceptions import ConnectionClosed


class SocketTransport:

    def __init__(self, sock: socket):
        self.sock = sock

    def write(self, data: bytes) -> None:
        """
        Write given bytes to socket. Loop as long as every byte in
        string is written unless exception is raised.
        """
        self.sock.sendall(data)

    def read(self, length: int) -> bytes:
        """
        Read as many bytes from socket as specified in length.
        Loop as long as every byte is read unless exception is raised.
        """
        data = bytearray()
        while len(data) != length:
            data += self.sock.recv((length - len(data)))
            if not data:
                raise ConnectionClosed('Connection unexpectedly closed.')
        return data

    def close(self) -> None:
        self.sock.close()



# -*- coding: UTF-8 -*-
import typing


class LibRouterosError(Exception):
    """Base exception for all other."""


class ConnectionClosed(LibRouterosError):
    """Raised when connection have been closed."""


class ProtocolError(LibRouterosError):
    """Raised when e.g. encoding/decoding fails."""


class FatalError(ProtocolError):
    """Exception raised when !fatal is received."""


class TrapError(ProtocolError):
    """
    Exception raised when !trap is received.

    :param int category: Optional integer representing category.
    :param str message: Error message.
    """

    def __init__(self, message: str, category: typing.Union[None, int] = None):
        self.category = category
        self.message = message
        super().__init__()

    def __str__(self) -> str:
        return str(self.message.replace('\r\n', ','))

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self})'


class MultiTrapError(ProtocolError):
    """
    Exception raised when multiple !trap words have been received in one response.

    :param traps: TrapError instances.
    """

    def __init__(self, *traps: TrapError):
        self.traps = traps
        super().__init__()

    def __str__(self) -> str:
        return ', '.join(str(trap) for trap in self.traps)



# -*- coding: UTF-8 -*-

import os
from setuptools import setup

here = os.path.dirname(__file__)


def read(fname):
    """
    Read given file's content.
    :param str fname: file name
    :returns: file contents
    :rtype: str
    """
    return open(os.path.join(here, fname)).read()


install_pkgs = (
        )

setup_pkgs = (
        'setuptools>=12.0.5',
        )

setup(
    install_requires=install_pkgs,
    setup_requires=setup_pkgs,
    zip_safe=False,
    name='librouteros',
    version='3.2.0',
    description='Python implementation of MikroTik RouterOS API',
    long_description=read('README.rst'),
    author='Łukasz Kostka',
    author_email='lukasz.kostka@netng.pl',
    url='https://github.com/luqasz/librouteros',
    packages=['librouteros'],
    license='GNU GPLv2',
    keywords='mikrotik routeros api',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: MacOS',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Software Development :: Libraries'
        ]
     )



# -*- coding: UTF-8 -*-

from typing import (
    Dict,
    Iterator,
    Any,
    List,
)

ReplyDict = Dict[str, Any]
ResponseIter = Iterator[ReplyDict]
QueryGen = Iterator[str]
Response = List[Dict[str, Any]]
